@@ -571,8 +571,8 @@ static int storage_write(void *storage, const int clsid, const int item_age) {
 
 static pthread_t storage_write_tid;
 static pthread_mutex_t storage_write_plock;
-#define WRITE_SLEEP_MAX 1000000
 #define WRITE_SLEEP_MIN 500
+#define MIN_PAGES_FREE 3
 
 static void *storage_write_thread(void *arg) {
     void *storage = arg;
@@ -592,17 +592,17 @@ static void *storage_write_thread(void *arg) {
     while (1) {
         // cache per-loop to avoid calls to the slabs_clsid() search loop
         int min_class = slabs_clsid(settings.ext_item_size);
+        unsigned int global_pages = global_page_pool_size(NULL);
         bool do_sleep = true;
         counter++;
-        if (to_sleep > WRITE_SLEEP_MAX)
-            to_sleep = WRITE_SLEEP_MAX;
+        if (to_sleep > settings.ext_max_sleep)
+            to_sleep = settings.ext_max_sleep;
 
         for (int x = 0; x < MAX_NUMBER_OF_SLAB_CLASSES; x++) {
             bool did_move = false;
             bool mem_limit_reached = false;
             unsigned int chunks_free;
             int item_age;
-            int target = settings.ext_free_memchunks[x];
             if (min_class > x || (backoff[x] && (counter % backoff[x] != 0))) {
                 // Long sleeps means we should retry classes sooner.
                 if (to_sleep > WRITE_SLEEP_MIN * 10)
@@ -611,13 +611,15 @@ static void *storage_write_thread(void *arg) {
             }
 
             // Avoid extra slab lock calls during heavy writing.
+            unsigned int chunks_perpage = 0;
             chunks_free = slabs_available_chunks(x, &mem_limit_reached,
-                    NULL);
+                    &chunks_perpage);
+            unsigned int target = chunks_perpage * MIN_PAGES_FREE;
 
             // storage_write() will fail and cut loop after filling write buffer.
             while (1) {
                 // if we are low on chunks and no spare, push out early.
-                if (chunks_free < target && mem_limit_reached) {
+                if (chunks_free < target && global_pages <= settings.ext_global_pool_min) {
                     item_age = 0;
                 } else {
                     item_age = settings.ext_item_age;
@@ -625,7 +627,6 @@ static void *storage_write_thread(void *arg) {
                 if (storage_write(storage, x, item_age)) {
                     chunks_free++; // Allow stopping if we've done enough this loop
                     did_move = true;
-                    do_sleep = false;
                     if (to_sleep > WRITE_SLEEP_MIN)
                         to_sleep /= 2;
                 } else {
@@ -636,7 +637,7 @@ static void *storage_write_thread(void *arg) {
             if (!did_move) {
                 backoff[x]++;
             } else if (backoff[x]) {
-                backoff[x] /= 2;
+                backoff[x] = 1;
             }
         }
 
@@ -644,7 +645,7 @@ static void *storage_write_thread(void *arg) {
         pthread_mutex_unlock(&storage_write_plock);
         if (do_sleep) {
             usleep(to_sleep);
-            to_sleep *= 2;
+            to_sleep++;
         }
         pthread_mutex_lock(&storage_write_plock);
     }
@@ -685,6 +686,7 @@ int start_storage_write_thread(void *arg) {
             strerror(ret));
         return -1;
     }
+    thread_setname(storage_write_tid, "mc-ext-write");
 
     return 0;
 }
@@ -765,7 +767,6 @@ static int storage_compact_check(void *storage, logger *l,
 static pthread_t storage_compact_tid;
 static pthread_mutex_t storage_compact_plock;
 #define MIN_STORAGE_COMPACT_SLEEP 10000
-#define MAX_STORAGE_COMPACT_SLEEP 2000000
 
 struct storage_compact_wrap {
     obj_io io;
@@ -888,7 +889,7 @@ static void _storage_compact_cb(void *e, obj_io *io, int ret) {
 // I guess it's only COLD. that's probably fine.
 static void *storage_compact_thread(void *arg) {
     void *storage = arg;
-    useconds_t to_sleep = MAX_STORAGE_COMPACT_SLEEP;
+    useconds_t to_sleep = settings.ext_max_sleep;
     bool compacting = false;
     uint64_t page_version = 0;
     uint64_t page_size = 0;
@@ -974,11 +975,11 @@ static void *storage_compact_thread(void *arg) {
             }
             pthread_mutex_unlock(&wrap.lock);
 
-            if (to_sleep > MIN_STORAGE_COMPACT_SLEEP)
-                to_sleep /= 2;
+            // finish actual compaction quickly.
+            to_sleep = MIN_STORAGE_COMPACT_SLEEP;
         } else {
-            if (to_sleep < MAX_STORAGE_COMPACT_SLEEP)
-                to_sleep += MIN_STORAGE_COMPACT_SLEEP;
+            if (to_sleep < settings.ext_max_sleep)
+                to_sleep += settings.ext_max_sleep;
         }
     }
     free(readback_buf);
@@ -1019,6 +1020,7 @@ int start_storage_compact_thread(void *arg) {
             strerror(ret));
         return -1;
     }
+    thread_setname(storage_compact_tid, "mc-ext-compact");
 
     return 0;
 }
@@ -1124,6 +1126,7 @@ void *storage_init_config(struct settings *s) {
     s->ext_wbuf_size = 1024 * 1024 * 4;
     s->ext_compact_under = 0;
     s->ext_drop_under = 0;
+    s->ext_max_sleep = 1000000;
     s->slab_automove_freeratio = 0.01;
     s->ext_page_size = 1024 * 1024 * 64;
     s->ext_io_threadcount = 1;
@@ -1155,6 +1158,7 @@ int storage_read_config(void *conf, char **subopt) {
         EXT_RECACHE_RATE,
         EXT_COMPACT_UNDER,
         EXT_DROP_UNDER,
+        EXT_MAX_SLEEP,
         EXT_MAX_FRAG,
         EXT_DROP_UNREAD,
         SLAB_AUTOMOVE_FREERATIO, // FIXME: move this back?
@@ -1172,6 +1176,7 @@ int storage_read_config(void *conf, char **subopt) {
         [EXT_RECACHE_RATE] = "ext_recache_rate",
         [EXT_COMPACT_UNDER] = "ext_compact_under",
         [EXT_DROP_UNDER] = "ext_drop_under",
+        [EXT_MAX_SLEEP] = "ext_max_sleep",
         [EXT_MAX_FRAG] = "ext_max_frag",
         [EXT_DROP_UNREAD] = "ext_drop_unread",
         [SLAB_AUTOMOVE_FREERATIO] = "slab_automove_freeratio",
@@ -1286,6 +1291,16 @@ int storage_read_config(void *conf, char **subopt) {
                 return 1;
             }
             break;
+        case EXT_MAX_SLEEP:
+            if (subopts_value == NULL) {
+                fprintf(stderr, "Missing ext_max_sleep argument\n");
+                return 1;
+            }
+            if (!safe_strtoul(subopts_value, &settings.ext_max_sleep)) {
+                fprintf(stderr, "could not parse argument to ext_max_sleep\n");
+                return 1;
+            }
+            break;
         case EXT_MAX_FRAG:
             if (subopts_value == NULL) {
                 fprintf(stderr, "Missing ext_max_frag argument\n");
@@ -1368,10 +1383,8 @@ void *storage_init(void *conf) {
         settings.ext_drop_under = cf->storage_file->page_count / 4;
     }
     crc32c_init();
-    /* Init free chunks to zero. */
-    for (int x = 0; x < MAX_NUMBER_OF_SLAB_CLASSES; x++) {
-        settings.ext_free_memchunks[x] = 0;
-    }
+
+    settings.ext_global_pool_min = 0;
     storage = extstore_init(cf->storage_file, ext_cf, &eres);
     if (storage == NULL) {
         fprintf(stderr, "Failed to initialize external storage: %s\n",
