@@ -53,6 +53,8 @@ struct conn_queue_item {
     enum conn_queue_item_modes mode;
     conn *c;
     void    *ssl;
+    uint64_t conntag;
+    enum protocol bproto;
     io_pending_t *io; // IO when used for deferred IO handling.
     STAILQ_ENTRY(conn_queue_item) i_next;
 };
@@ -385,6 +387,8 @@ static void create_worker(void *(*func)(void *), void *arg) {
                 strerror(ret));
         exit(1);
     }
+
+    thread_setname(((LIBEVENT_THREAD*)arg)->thread_id, "mc-worker");
 }
 
 /*
@@ -488,7 +492,7 @@ static void setup_thread(LIBEVENT_THREAD *me) {
     // TODO: maybe register hooks to be called here from sub-packages? ie;
     // extstore, TLS, proxy.
     if (settings.proxy_enabled) {
-        proxy_thread_init(me);
+        proxy_thread_init(settings.proxy_ctx, me);
     }
 #endif
     thread_io_queue_add(me, IO_QUEUE_NONE, NULL, NULL, NULL, NULL, NULL);
@@ -567,7 +571,7 @@ static void thread_libevent_process(evutil_socket_t fd, short which, void *arg)
             case queue_new_conn:
                 c = conn_new(item->sfd, item->init_state, item->event_flags,
                                    item->read_buffer_size, item->transport,
-                                   me->base, item->ssl);
+                                   me->base, item->ssl, item->conntag, item->bproto);
                 if (c == NULL) {
                     if (IS_UDP(item->transport)) {
                         fprintf(stderr, "Can't listen for events on UDP socket\n");
@@ -627,6 +631,17 @@ static void thread_libevent_process(evutil_socket_t fd, short which, void *arg)
     }
 }
 
+// Interface is slightly different on various platforms.
+// On linux, at least, the len limit is 16 bytes.
+#define THR_NAME_MAXLEN 16
+void thread_setname(pthread_t thread, const char *name) {
+assert(strlen(name) < THR_NAME_MAXLEN);
+#if defined(__linux__)
+pthread_setname_np(thread, name);
+#endif
+}
+#undef THR_NAME_MAXLEN
+
 // NOTE: need better encapsulation.
 // used by the proxy module to iterate the worker threads.
 LIBEVENT_THREAD *get_worker_thread(int id) {
@@ -713,7 +728,8 @@ select:
  * of an incoming connection.
  */
 void dispatch_conn_new(int sfd, enum conn_states init_state, int event_flags,
-                       int read_buffer_size, enum network_transport transport, void *ssl) {
+                       int read_buffer_size, enum network_transport transport, void *ssl,
+                       uint64_t conntag, enum protocol bproto) {
     CQ_ITEM *item = NULL;
     LIBEVENT_THREAD *thread;
 
@@ -737,6 +753,8 @@ void dispatch_conn_new(int sfd, enum conn_states init_state, int event_flags,
     item->transport = transport;
     item->mode = queue_new_conn;
     item->ssl = ssl;
+    item->conntag = conntag;
+    item->bproto = bproto;
 
     MEMCACHED_CONN_DISPATCH(sfd, (int64_t)thread->thread_id);
     notify_worker(thread, item);
@@ -913,11 +931,11 @@ enum store_item_type store_item(item *item, int comm, conn* c) {
 
 /******************************* GLOBAL STATS ******************************/
 
-void STATS_LOCK() {
+void STATS_LOCK(void) {
     pthread_mutex_lock(&stats_lock);
 }
 
-void STATS_UNLOCK() {
+void STATS_UNLOCK(void) {
     pthread_mutex_unlock(&stats_lock);
 }
 
